@@ -8,6 +8,7 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <ctype.h>
+#include <endian.h>
 #include <sys/ioctl.h>
 #include <sys/shm.h>
 #include <sys/stat.h>
@@ -19,7 +20,7 @@
 #include <linux/major.h>
 #include <sys/mount.h>
 
-#define VERSION "v4.1"
+#define VERSION "v4.5"
 
 #undef	 DO_FLUSHCACHE		/* under construction: force cache flush on -W0 */
 
@@ -98,6 +99,11 @@ static int	hwif_data = 0;
 static int	hwif_ctrl = 0;
 static int	hwif_irq = 0;
 #endif
+#ifdef HDIO_GET_BUSSTATE
+static int	set_busstate = 0, get_busstate = 0, busstate = 0;
+#endif
+
+extern void identify (void *, const char *);
 
 #ifdef HDIO_DRIVE_RESET
 static int	perform_reset = 0;
@@ -163,7 +169,7 @@ char *BuffType[]	= {"unknown", "1Sect", "DualPort", "DualPortCache"};
 static void dmpstr (char *prefix, unsigned int i, char *s[], unsigned int maxi)
 {
 	if (i > maxi)
-		printf("%s%d", prefix, i);
+		printf("%s%u", prefix, i);
 	else
 		printf("%s%s", prefix, s[i]);
 }
@@ -182,17 +188,17 @@ static void dump_identity (struct hd_driveid *id)
 			printf("%s", cfg_str[i]);
 	}
 	printf(" }\n");
-	printf(" RawCHS=%d/%d/%d, TrkSize=%d, SectSize=%d, ECCbytes=%d\n",
+	printf(" RawCHS=%u/%u/%u, TrkSize=%u, SectSize=%u, ECCbytes=%u\n",
 		id->cyls, id->heads, id->sectors,
 		id->track_bytes, id->sector_bytes, id->ecc_bytes);
 	dmpstr (" BuffType=",id->buf_type,BuffType,3);
-	printf(", BuffSize=%dkB, MaxMultSect=%d", id->buf_size/2, id->max_multsect);
+	printf(", BuffSize=%ukB, MaxMultSect=%u", id->buf_size/2, id->max_multsect);
 	if (id->max_multsect) {
 		printf(", MultSect=");
 		if (!(id->multsect_valid&1))
-			printf("?%d?", id->multsect);
+			printf("?%u?", id->multsect);
 		else if (id->multsect)
-			printf("%d", id->multsect);
+			printf("%u", id->multsect);
 		else
 			printf("off");
 	}
@@ -204,17 +210,20 @@ static void dump_identity (struct hd_driveid *id)
 	}
 	if (!(id->field_valid&1))
 		printf(" (maybe):");
-	printf(" CurCHS=%d/%d/%d, CurSects=%d",
+#if __BYTE_ORDER == __BIG_ENDIAN
+	printf(" CurCHS=%u/%u/%u, CurSects=%u",
 		id->cur_cyls, id->cur_heads, id->cur_sectors,
-#ifdef BIG_ENDIAN
 		(id->cur_capacity0 << 16) | id->cur_capacity1
-#else
-		(id->cur_capacity1 << 16) | id->cur_capacity0
-#endif
 		);
+#else
+	printf(" CurCHS=%u/%u/%u, CurSects=%u",
+		id->cur_cyls, id->cur_heads, id->cur_sectors,
+		(id->cur_capacity1 << 16) | id->cur_capacity0
+		);
+#endif
 	printf(", LBA=%s", YN(id->capability&2));
 	if (id->capability&2)
- 		printf(", LBAsects=%d", id->lba_capacity);
+ 		printf(", LBAsects=%u", id->lba_capacity);
 
 	if (id->capability&1) {
 		if (id->dma_1word | id->dma_mword) {
@@ -243,7 +252,7 @@ static void dump_identity (struct hd_driveid *id)
 		printf("no");
 	if ((id->capability&8) || (id->field_valid&2)) {
 		if (id->field_valid&2) {
-			printf(", tPIO={min:%d,w/IORDY:%d}", id->eide_pio, id->eide_pio_iordy);
+			printf(", tPIO={min:%u,w/IORDY:%u}", id->eide_pio, id->eide_pio_iordy);
 			if (id->eide_pio_modes & 1)	strcat(pmodes, "pio3 ");
 			if (id->eide_pio_modes & 2)	strcat(pmodes, "pio4 ");
 			if (id->eide_pio_modes &~3)	strcat(pmodes, "pio? ");
@@ -280,7 +289,7 @@ static void dump_identity (struct hd_driveid *id)
 		}
 	}
 	if ((id->capability&1) && (id->field_valid&2))
-		printf(", tDMA={min:%d,rec:%d}", id->eide_dma_min, id->eide_dma_time);
+		printf(", tDMA={min:%u,rec:%u}", id->eide_dma_min, id->eide_dma_time);
 	printf("\n PIO modes: %s", pmodes);
 	printf("\n DMA modes: %s", dmodes);
 
@@ -291,8 +300,10 @@ static void dump_identity (struct hd_driveid *id)
 		else if ((id_regs[91]&0xFF00)!=0x4000) 
 			printf(": unknown setting"); 
 		else 
-			printf(": mode=0x%02X (%d)",id_regs[91]&0xFF,id_regs[91]&0xFF); 
+			printf(": mode=0x%02X (%u)",id_regs[91]&0xFF,id_regs[91]&0xFF); 
 	} 
+	if (id_regs[82]&0x20)
+		printf(" WriteCache=%s",(id_regs[85]&0x20) ? "enabled" : "disabled");
 #ifdef __NEW_HD_DRIVE_ID
 	if ((id->minor_rev_num) || (id->major_rev_num)) {
 		printf("\n Drive Supports : ");
@@ -526,6 +537,29 @@ static void on_off (unsigned int value)
 	printf(value ? " (on)\n" : " (off)\n");
 }
 
+#ifdef HDIO_GET_BUSSTATE
+static void bus_state_value (unsigned int value)
+{
+	char *string;
+
+	switch (value) {
+	case BUSSTATE_ON:
+		string = " (on)\n";
+		break;
+	case BUSSTATE_OFF:
+		string = " (off)\n";
+		break;
+	case BUSSTATE_TRISTATE:
+		string = " (tristate)\n";
+		break;
+	default:
+		string = " (unknown: %d)\n";
+		break;
+	}
+	printf(string, value);
+}
+#endif
+
 #ifdef HDIO_DRIVE_CMD
 static void interpret_standby (unsigned int standby)
 {
@@ -546,16 +580,16 @@ static void interpret_standby (unsigned int standby)
 				unsigned int secs = standby * 5;
 				unsigned int mins = secs / 60;
 				secs %= 60;
-				if (mins)	  printf("%d minutes", mins);
+				if (mins)	  printf("%u minutes", mins);
 				if (mins && secs) printf(" + ");
-				if (secs)	  printf("%d seconds", secs);
+				if (secs)	  printf("%u seconds", secs);
 			} else if (standby <= 251) {
 				unsigned int mins = (standby - 240) * 30;
 				unsigned int hrs  = mins / 60;
 				mins %= 60;
-				if (hrs)	  printf("%d hours", hrs);
+				if (hrs)	  printf("%u hours", hrs);
 				if (hrs && mins)  printf(" + ");
-				if (mins)	  printf("%d minutes", mins);
+				if (mins)	  printf("%u minutes", mins);
 			} else
 				printf("illegal value)\n");
 			break;
@@ -578,7 +612,7 @@ static void interpret_xfermode (unsigned int xfermode)
 		case 12:
 		case 13:
 		case 14:
-		case 15:	printf("PIO flow control mode%d", xfermode-8);
+		case 15:	printf("PIO flow control mode%u", xfermode-8);
 				break;
 		case 16:
 		case 17:
@@ -587,7 +621,7 @@ static void interpret_xfermode (unsigned int xfermode)
 		case 20:
 		case 21:
 		case 22:
-		case 23:	printf("singleword DMA mode%d", xfermode-16);
+		case 23:	printf("singleword DMA mode%u", xfermode-16);
 				break;
 		case 32:
 		case 33:
@@ -596,7 +630,7 @@ static void interpret_xfermode (unsigned int xfermode)
 		case 36:
 		case 37:
 		case 38:
-		case 39:	printf("multiword DMA mode%d", xfermode-32);
+		case 39:	printf("multiword DMA mode%u", xfermode-32);
 				break;
 		case 64:
 		case 65:
@@ -605,7 +639,7 @@ static void interpret_xfermode (unsigned int xfermode)
 		case 68:
 		case 69:
 		case 70:
-		case 71:	printf("UltraDMA mode%d", xfermode-64);
+		case 71:	printf("UltraDMA mode%u", xfermode-64);
 				break;
 		default:
 				printf("unknown, probably not valid");
@@ -707,7 +741,7 @@ void process_dev (char *devname)
 #ifdef HDIO_UNREGISTER_HWIF
 	if (unregister_hwif) {
 		no_scsi();
-		printf(" attempting to unregister hwif#%d\n", hwif);
+		printf(" attempting to unregister hwif#%u\n", hwif);
 		if (ioctl(fd, HDIO_UNREGISTER_HWIF, hwif)) 
 			perror(" HDIO_UNREGISTER_HWIF failed");
 	}
@@ -716,7 +750,7 @@ void process_dev (char *devname)
 	if (scan_hwif) {
 		int	args[3];
 		no_scsi();
-		printf(" attempting to scan hwif (0x%x, 0x%x, %d)\n", hwif_data, hwif_ctrl, hwif_irq);
+		printf(" attempting to scan hwif (0x%x, 0x%x, %u)\n", hwif_data, hwif_ctrl, hwif_irq);
 		args[0] = hwif_data;
 		args[1] = hwif_ctrl;
 		args[2] = hwif_irq;
@@ -1025,7 +1059,7 @@ void process_dev (char *devname)
 					break; 
 				case 3:	printf("32-bit w/sync)\n");
 					break; 
-				default:printf("???)\n");
+				default:printf("\?\?\?)\n");
 			}
 		}
 	}
@@ -1093,22 +1127,21 @@ void process_dev (char *devname)
 		}
 	}
 	if (verbose || get_geom) {
-#if HDIO_GETGEO_BIG
-		static struct hd_big_geometry g;
-#else
+		static const char msg[] = " geometry     = %u/%u/%u, sectors = %ld, start = %ld\n";
 		static struct hd_geometry g;
+#if HDIO_GETGEO_BIG
+		static struct hd_big_geometry bg;
 #endif
+
 		if (ioctl(fd, BLKGETSIZE, &parm))
 			perror(" BLKGETSIZE failed");
 #if HDIO_GETGEO_BIG
-		else if (ioctl(fd, HDIO_GETGEO_BIG, &g))
-			perror(" HDIO_GETGEO_BIG failed");
-#else
+		else if (!ioctl(fd, HDIO_GETGEO_BIG, &bg))
+			printf(msg, bg.cylinders, bg.heads, bg.sectors, parm, bg.start);
+#endif
 		else if (ioctl(fd, HDIO_GETGEO, &g))
 			perror(" HDIO_GETGEO failed");
-#endif
-		else	printf(" geometry     = %d/%d/%d, sectors = %ld, start = %ld\n",
-				g.cylinders, g.heads, g.sectors, parm, g.start);
+		else	printf(msg, g.cylinders, g.heads, g.sectors, parm, g.start);
 	}
 #ifdef HDIO_DRIVE_CMD
 	if (get_powermode) {
@@ -1177,17 +1210,37 @@ void process_dev (char *devname)
 		no_scsi();
 		no_xt();
 		if (!(ioctl(fd, HDIO_DRIVE_CMD, &args))) {
-			old_identity = 0;
-			dump_identity((struct hd_driveid *)&args[4]);
+			identify((void *)&args[4], NULL);
 		} else {
 			args[0] = WIN_PIDENTIFY;
 			if (!(ioctl(fd, HDIO_DRIVE_CMD, &args))) {
-				old_identity = 0;
-				dump_identity((struct hd_driveid *)&args[4]);
+				identify((void *)&args[4], NULL);
 			} else
 				perror(" HDIO_DRIVE_CMD(identify) failed");
 		}
 	}
+#endif
+#ifdef HDIO_SET_BUSSTATE
+	if (set_busstate) {
+		no_scsi();
+		if (get_busstate) {
+			printf(" setting bus state to %d", busstate);
+			bus_state_value(busstate);
+		}
+		if (ioctl(fd, HDIO_SET_BUSSTATE, busstate))
+			perror(" HDIO_SET_BUSSTATE failed");
+	}
+#endif
+#ifdef HDIO_GET_BUSSTATE
+	if ((verbose && !is_scsi_hd) || get_busstate) {
+		no_scsi();
+		if (ioctl(fd, HDIO_GET_BUSSTATE, &parm))
+			perror(" HDIO_GET_BUSSTATE failed");
+		else {
+			printf(" busstate     = %2ld", parm);
+			bus_state_value(parm);
+		}
+	}
 #endif
 	if (do_ctimings)
 		time_cache (fd);
@@ -1206,6 +1259,9 @@ void usage_error (void)
 	" -a   get/set fs readahead\n"
 #ifdef HDIO_DRIVE_CMD
 	" -A   set drive read-lookahead flag (0/1)\n"
+#endif
+#ifdef HDIO_GET_BUSSTATE
+	" -b   get/set bus state (0 == off, 1 == on, 2 == tristate)\n"
 #endif
 	" -B   get Advanced Power Management setting (1-255)\n"
 #ifdef HDIO_SET_32BIT
@@ -1220,13 +1276,13 @@ void usage_error (void)
 	" -D   enable/disable drive defect-mgmt\n"
 #ifdef CDROM_SELECT_SPEED
 	" -E   set cd-rom drive speed\n"
-#endif CDROM_SELECT_SPEED
+#endif
 	" -f   flush buffer cache for device on exit\n"
 	" -g   display drive geometry\n"
 	" -h   display terse usage information\n"
 	" -i   display drive identification\n"
 #ifdef HDIO_SET_32BIT
-	" -I   read drive identification directly from drive\n"
+	" -I   detailed/current information directly from drive\n"
 #endif
 #ifdef HDIO_GET_KEEPSETTINGS
 	" -k   get/set keep_settings_over_reset flag (0/1)\n"
@@ -1609,6 +1665,24 @@ int main(int argc, char **argv)
 						do_ctimings = 1;
 						do_flush = 1;
 						break;
+#ifdef HDIO_GET_BUSSTATE
+					case 'b':
+                                                get_busstate = noisy;
+                                                noisy = 1;
+                                                if (!*p && argc && isdigit(**argv))
+						  p = *argv++, --argc;
+						switch (*p) {
+						case '0':
+						case '1':
+						case '2':
+						  set_busstate = 1;
+						  busstate = *p++ - '0';
+						  break;
+						default:
+						  break;
+                                                }
+						break;
+#endif
 					case 'h':
 					default:
 						usage_error();
